import sqlite3
import pytest
from faker import Faker
from pluggy import PluginManager
from src.chercher.db import init_db
from src.chercher import hookspecs

fake = Faker()


@pytest.fixture
def plugin_manager():
    pm = PluginManager("chercher")
    pm.add_hookspecs(hookspecs)

    return pm


@pytest.fixture
def test_db():
    conn = sqlite3.connect(":memory:")
    init_db(conn)

    yield conn
    conn.close()



import sys
from loguru import logger
import click
from chercher.utils import console
from chercher.output import print_plugins_table, print_results_list, print_results_table
from chercher.plugin_manager import get_plugin_manager
from chercher.settings import settings, APP_NAME, APP_DIR, CONFIG_FILE_PATH
from chercher.app import ChercherApp
from chercher.db import init_db, db_connection
from chercher.db_actions import index, prune, search

logger.remove()
logger.add(
    sys.stderr,
    colorize=True,
    format="<green>{time:HH:mm:ss}</green> | <level>{level}</level> | {message}",
    level="INFO",
)
logger.add(
    APP_DIR / "chercher_errors.log",
    rotation="10 MB",
    retention="15 days",
    level="ERROR",
)


@click.group(help=settings.description)
@click.version_option(
    version=settings.version,
    message="v%(version)s",
    package_name=APP_NAME,
)
@click.pass_context
def cli(ctx: click.Context) -> None:
    with db_connection(settings.db_url) as conn:
        logger.debug("initializing the database")
        init_db(conn)

    ctx.ensure_object(dict)
    ctx.obj["settings"] = settings
    ctx.obj["db_url"] = settings.db_url
    ctx.obj["pm"] = get_plugin_manager()


@cli.command(
    name="index",
    help="Index documents, webpages and more.",
)
@click.argument("uris", nargs=-1)
@click.pass_context
def index_cmd(ctx: click.Context, uris: list[str]) -> None:
    pm = ctx.obj["pm"]
    db_url = ctx.obj["db_url"]

    if not pm.list_name_plugin():
        logger.warning("No plugins registered!")
        return

    with db_connection(db_url) as conn:
        index(conn, uris, pm)


@cli.command(
    name="prune",
    help="Prune unnecessary documents from the database.",
)
@click.pass_context
def prune_cmd(ctx: click.Context) -> None:
    pm = ctx.obj["pm"]
    db_url = ctx.obj["db_url"]

    with db_connection(db_url) as conn:
        prune(conn, pm)


@cli.command(
    name="search",
    help="Seach for documents matching your query.",
)
@click.argument("query")
@click.option(
    "-l",
    "--limit",
    type=int,
    default=5,
    help="Number of results.",
)
@click.option(
    "-o",
    "--output",
    type=click.Choice(["table", "list"], case_sensitive=False),
    default="table",
    help="Output format (available options: table and list).",
)
@click.pass_context
def search_cmd(
    ctx: click.Context, query: str, limit: int, output: str = "table"
) -> None:
    db_url = ctx.obj["db_url"]

    with db_connection(db_url) as conn:
        results = search(conn, query, limit)
        if not results:
            console.print(f"No results found for: '{query}'")
            return

        if output == "list":
            print_results_list(results)
        else:
            print_results_table(results)


@cli.command(help="List out all the registered plugins and their hooks.")
@click.pass_context
def plugins(ctx: click.Context) -> None:
    pm = ctx.obj["pm"]
    print_plugins_table(pm)


@cli.command(
    help="Print the location of the configuration file or the database.",
)
@click.argument(
    "item",
    type=click.Choice(["config", "db"]),
)
def locate(item: str) -> None:
    if item == "config":
        console.print(
            f"config file located at: [url={CONFIG_FILE_PATH.absolute()}]{CONFIG_FILE_PATH.absolute()}[/]"
        )
    elif item == "db":
        console.print(f"db located at: [url={settings.db_url}]{settings.db_url}[/]")


@cli.command(
    help="Starts the TUI.",
)
def app() -> None:
    app = ChercherApp()
    app.run()


if __name__ == "__main__":
    cli()



from datetime import datetime, timedelta
from importlib.metadata import version
from textual import on
from textual.app import App, ComposeResult
from textual.widgets import (
    Footer,
    Label,
    Input,
    Button,
    Link,
    ListItem,
    ListView,
    Tree,
)
from textual.containers import Container
from textual.validation import Number
from textual.reactive import reactive
from chercher.settings import APP_NAME, APP_DIR, Settings
from chercher.db import db_connection
from chercher.db_actions import search, load_history

settings = Settings()

# TODO: add notifications.
# TODO: signal input errors.
# TODO: enter to open link.
# TODO: improve tab navigation (?).
# TODO: add loading indicator.


def format_date(date_str: str) -> str:
    date = datetime.strptime(date_str, "%Y-%m-%d").date()
    today = datetime.today().date()
    yesterday = today - timedelta(days=1)

    if date == today:
        return "today"
    elif date == yesterday:
        return "yesterday"
    else:
        return date_str


class ChercherApp(App):
    TITLE = APP_NAME
    ENABLE_COMMAND_PALETTE = False
    BINDINGS = [
        ("ctrl+q", "quit", "quit"),
    ]

    CSS_PATH = "styles.tcss"

    search_query: reactive[str] = reactive("")
    n_results: reactive[int] = reactive(10)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.theme = settings.theme

    def on_mount(self):
        self.load_history()

    def compose(self) -> ComposeResult:
        with Container(classes="header"):
            yield Label(f"↪ {APP_NAME}", classes="title")
            yield Label(f"v{version(APP_NAME)}", classes="version")
            yield Label(f"{APP_DIR}", classes="path")

        with Container(classes="search-bar"):
            yield Input(
                placeholder="search...",
                type="text",
                classes="input input--query",
            )
            yield Input(
                placeholder="10",
                type="integer",
                validators=Number(minimum=1, maximum=100),
                classes="input input--n-results",
            )
            yield Button(label="search", classes="submit")

        with Container(classes="main") as _:
            with Tree(classes="history", label="*") as history:
                history.show_root = False
                history.show_guides = True
                history.root.expand()

                history.border_title = "history"

            with ListView(classes="results") as results:
                results.border_title = "results"

        yield Footer()

    @on(Input.Changed, selector=".input--query")
    def update_search_query(self, event: Input.Changed) -> None:
        self.search_query = event.value

    @on(Input.Changed, selector=".input--n-results")
    def update_number_of_results(self, event: Input.Changed) -> None:
        try:
            self.n_results = int(event.value)
        except Exception:
            self.n_results = 1

    @on(Input.Submitted)
    @on(Button.Pressed)
    def submit(self) -> None:
        if not self.search_query:
            return

        results_list: ListView = self.query_one(".results")
        results_list.clear()
        results_list.loading = True

        with db_connection(settings.db_url) as conn:
            results = search(conn, self.search_query, self.n_results)
            if not results:
                self.notify(f'no results found for "{self.search_query}"')
                results_list.loading = False
                return

            for result in results:
                results_list.mount(
                    ListItem(
                        Link(
                            result[1] or result[0],
                            url=result[0],
                            tooltip="click to open",
                        ),
                        classes="result",
                    ),
                )

        self.load_history()
        results_list.loading = False

    @on(Tree.NodeSelected)
    def search_from_history(self, event: Tree.NodeSelected) -> None:
        if event.node.data and event.node.data.get("is_root", False):
            return

        search_input: Input = self.query_one(".input--query")
        self.search_query = f"{event.node.label}"
        search_input.value = self.search_query
        search_input.focus(True)

    def load_history(self) -> None:
        history_tree: Tree = self.query_one(".history")
        history_tree.clear()
        history_tree.loading = True

        with db_connection(settings.db_url) as conn:
            entries = load_history(conn)
            if not entries:
                history_tree.loading = False
                return

            for date, queries in entries:
                date_str = format_date(date)
                today_or_yesterday = date_str in ("today", "yesterday")
                date_root = history_tree.root.add(
                    f"{format_date(date)}",
                    expand=today_or_yesterday,
                    data={"is_root": True},
                )
                for query in queries.split(","):
                    date_root.add_leaf(query)

            history_tree.loading = False


if __name__ == "__main__":
    app = ChercherApp()
    app.run()



from src.chercher.db_actions import index
from .plugin_mocks import DummyTxtPlugin, BadPlugin


def test_index_valid_txt(faker, test_db, plugin_manager):
    plugin_manager.register(DummyTxtPlugin())
    uri = faker.file_path(extension="txt")
    index(test_db, [uri], plugin_manager)

    cursor = test_db.cursor()
    cursor.execute("SELECT * FROM documents")
    documents = cursor.fetchall()

    assert len(documents) == 1
    assert documents[0][0] == uri
    assert documents[0][1] is not None
    assert documents[0][2] is not None
    assert documents[0][3] is not None


def test_index_with_exception(faker, test_db, plugin_manager):
    plugin_manager.register(DummyTxtPlugin())
    plugin_manager.register(BadPlugin())
    uris = [
        faker.file_path(depth=3, extension="pdf"),
        faker.file_path(depth=3, extension="txt"),
        faker.file_path(depth=3, extension="epub"),
    ]

    index(test_db, uris, plugin_manager)

    cursor = test_db.cursor()
    cursor.execute("SELECT * FROM documents")
    documents = cursor.fetchall()

    assert len(documents) == 1


def test_index_same_document_multiple_times(faker, test_db, plugin_manager):
    plugin_manager.register(DummyTxtPlugin())
    uri = faker.file_path(depth=3, extension="txt")

    index(test_db, [uri, uri], plugin_manager)

    cursor = test_db.cursor()
    cursor.execute("SELECT * FROM documents WHERE uri = ?", (uri,))
    documents = cursor.fetchall()

    assert len(documents) == 1
    assert documents[0][0] == uri



from typing import Any
import sqlite3
from loguru import logger
import pluggy
from rich.progress import Progress
from chercher.settings import settings


def index(conn: sqlite3.Connection, uris: list[str], pm: pluggy.PluginManager) -> None:
    cursor = conn.cursor()
    plugin_settings = dict(settings).get("plugin", {})

    for uri in uris:
        try:
            for documents in pm.hook.ingest(uri=uri, settings=plugin_settings):
                for doc in documents:
                    try:
                        cursor.execute(
                            """
                    INSERT INTO documents (uri, title, body, hash, metadata) VALUES (?, ?, ?, ?, ?)
                    """,
                            (
                                doc.uri,
                                doc.title,
                                doc.body,
                                doc.hash,
                                doc.metadata.model_dump_json(),
                            ),
                        )
                        conn.commit()
                        logger.info(f'document "{doc.uri}" indexed')
                    except sqlite3.IntegrityError:
                        logger.warning(f'document "{doc.uri}" already exists')
                    except Exception as e:
                        logger.error(
                            f"something went wrong while indexing '{doc.uri}': {e}"
                        )
        except Exception as e:
            logger.error(
                f"something went wrong while trying to index documents from '{uri}': {e}"
            )


def prune(conn: sqlite3.Connection, pm: pluggy.PluginManager) -> None:
    cursor = conn.cursor()
    plugin_settings = dict(settings).get("plugin", {})
    with Progress(transient=True) as progress:
        task = progress.add_task(description="getting documents...", total=1)

        try:
            cursor.execute("SELECT uri, hash FROM documents")
            uris_and_hashes = cursor.fetchall()
        except Exception as e:
            logger.error(
                f"something went wrong while retrieving documents from the database: {e}"
            )
            return

        progress.update(task, advance=0.25, description="pruning documents...")
        for uri, hash in uris_and_hashes:
            try:
                for result in pm.hook.prune(
                    uri=uri, hash=hash, settings=plugin_settings
                ):
                    if not result:
                        continue

                    try:
                        cursor.execute("DELETE FROM documents WHERE uri = ?", (uri,))
                        conn.commit()
                        logger.info(f"document '{uri}' pruned")
                    except Exception as e:
                        logger.error(f"something went wrong while purging '{uri}': {e}")
            except Exception as e:
                logger.error(
                    f"something went wrong while trying to purge document '{uri}': {e}"
                )

        try:
            progress.update(task, advance=0.25, description="cleaning up database...")
            cursor.execute("VACUUM;")
            cursor.execute("PRAGMA optimize;")
        except Exception as e:
            logger.error(f"something went wrong while performing vacuum operation: {e}")
            return

        progress.update(task, completed=True)
    logger.info("database cleanup completed")


def search(conn: sqlite3.Connection, query: str, limit: int) -> list[Any] | None:
    cursor = conn.cursor()

    cursor.execute(
        """
            INSERT INTO history (query, created_at, last_updated_at)
            VALUES (?, CURRENT_TIMESTAMP, CURRENT_TIMESTAMP)
            ON CONFLICT(query) DO UPDATE SET
            last_updated_at = CURRENT_TIMESTAMP;
        """,
        (query,),
    )

    cursor.execute(
        """
            SELECT uri, title, substr(body, 0, 300)
            FROM documents
            WHERE ROWID IN (
                SELECT ROWID
                FROM documents_fts
                WHERE documents_fts MATCH ?
                ORDER BY bm25(documents_fts)
                LIMIT ?
            )
            """,
        (query, limit),
    )
    results = cursor.fetchall()
    return results


def load_history(conn: sqlite3.Connection, limit: int = 1_000) -> list[Any] | None:
    cursor = conn.cursor()

    cursor.execute(
        """
        SELECT DATE(last_updated_at) AS date, GROUP_CONCAT(query) AS queries
        FROM history
        GROUP BY date
        ORDER BY date DESC
        LIMIT ?;
        """,
        (limit,),
    )

    results = cursor.fetchall()
    return results



import textwrap
from pluggy import PluginManager
from rich.table import Table
from rich.progress import track
from chercher.utils import console


def print_results_table(results: list[dict] = []) -> None:
    table = Table(
        title="results",
        show_lines=True,
        padding=1,
    )
    table.add_column("title")
    table.add_column("summary")

    for result in results:
        title = result[1] if result[1] else result[0]
        table.add_row(
            f"[link={result[0]}]{title}[/]",
            f"{textwrap.shorten(result[2], width=280, placeholder='...')}",
        )

    console.print(table)


def print_results_list(results: list[dict] = []) -> None:
    grid = Table(
        title="results",
        expand=True,
        show_lines=False,
        show_footer=False,
        show_header=False,
        show_edge=False,
    )
    grid.add_column(justify="left")

    for result in results:
        uri = result[0]
        title = f"[link={uri}][bold]{result[1] if result[1] else result[0]}[/]"
        summary = f"{textwrap.shorten(result[2], width=280, placeholder='...')}\n"

        grid.add_row(f"{title}\n[underline italic]{uri}[/]\n{summary}")

    console.print(grid)


def print_plugins_table(pm: PluginManager) -> None:
    table = Table(title="plugins")
    table.add_column("name")
    table.add_column("version")
    table.add_column("hooks")

    plugins = dict(pm.list_plugin_distinfo())
    for plugin, dist_info in track(plugins.items(), description="checking plugins..."):
        version = f"v{dist_info.version}" if dist_info else "n/a"
        hooks = [h.name for h in pm.get_hookcallers(plugin)]
        hooks_str = ", ".join(hooks) if hooks else ""
        table.add_row(plugin.__name__, version, hooks_str)

    console.clear()
    console.print(table)



from .plugin_mocks import DummyPlugin


def test_dummy_plugin_yields_document_correctly(faker, plugin_manager):
    plugin_manager.register(DummyPlugin())
    uri = faker.file_path(extension="txt")
    documents = [
        document
        for doc_gen in plugin_manager.hook.ingest(uri=uri)
        for document in doc_gen
    ]
    assert len(documents) == 1
    assert documents[0].uri == uri


def test_dummy_plugin_yields_prune_result_correctly(faker, plugin_manager):
    plugin_manager.register(DummyPlugin())
    results = [result for result in plugin_manager.hook.prune(uri=faker.file_path())]
    assert len(results) == 1
    assert results[0]
